/**====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*
 *
 * IOTCAPIs_WakeUp_Device.c
 *
 * Copyright (c) by TUTK Co.LTD. All Rights Reserved.
 *
 * @brief   Examle of how to use IOTC WakeUp API at Device side
 *
 *====*====*====*====*====*====*====*====*====*====*====*====*====*====*====*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include "IOTCDevice.h"
#include "IOTCWakeUp.h"
#include "sample_constants.h"


// Macro definition
#define MAX_BUF_SIZE 1500
#define ERROR -1
#define NOT_READY 0
#define READY 1

// Structure declaration
typedef struct _WakeupParam
{
    int                 nSkt;
    unsigned long       ulAliveAddr;
    unsigned short      usAlivePort;
    char                *pWakeupPattern;
    int                 nWakeupPatternLength;
    char                *pWakeupResponse;
    int                 nWakeupResponseLength;
} WakeupParam;


typedef struct _AliveParam
{
    int 			  nSkt;
    IOTC_WakeUpDataEx *pWUData;
    int 			  nDataCnt;
} AliveParam;

typedef enum _SLEEP_PROTOCOL{
    PROTOCOL_UDP,
    PROTOCOL_TCP
} SLEEP_PROTOCOL;


// Variables
static bool g_bWakeUp = false;
static bool g_bConnected = true;
static bool g_deviceLogined = false;
IOTC_WakeUpDataEx *gAliveData = NULL;
unsigned int gDataCnt = 0;
int gError;
SLEEP_PROTOCOL gSleepProtocol;

static int createSocket()
{
    int nSkt = -1;
    if(gSleepProtocol == PROTOCOL_UDP){
        nSkt = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    }
    else if(gSleepProtocol == PROTOCOL_TCP){
        nSkt = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    }
    printf("create socket = %d!\n", nSkt);

    return nSkt;
}


static void SetupSocketSelectParam(int nSkt, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *time, uint8_t nMaxTimeout)
{
    if(readfds)
    {
        FD_ZERO(readfds);
        FD_SET(nSkt, readfds);
    }

    if(writefds)
    {
        FD_ZERO(writefds);
        FD_SET(nSkt, writefds);
    }

    if(exceptfds)
    {
        FD_ZERO(exceptfds);
        FD_SET(nSkt, exceptfds);
    }

    if(time)
    {
        memset(time, 0, sizeof(struct timeval));
        time->tv_sec = nMaxTimeout;
    }
}


static bool isWakeupPattern(const char *pData, const char *pPattern, const int nPatternLength)
{
    bool bPattern = true;
    for(int i = 0; i < nPatternLength; i++)
    {
        if(pPattern[i]){
            if(pData[i] != pPattern[i])
            {
                bPattern = false;
                break;
            }
        }
    }

    return bPattern;
}

static void sendWakeupResponse(int nSkt, char *msg, int len);
static void packetParser(WakeupParam *pParam)
{
    char buf[MAX_BUF_SIZE];
    int nRecvSize = 0;

    socklen_t socklength = 0;
    int nSkt = pParam->nSkt;

    //printf("[packetParser] - Prepare to receive socket header. skt = %d\n", nSkt);

    struct sockaddr_in remote;
    memset(&remote, 0, sizeof(struct sockaddr_in));
    remote.sin_family = AF_INET;
    remote.sin_addr.s_addr = pParam->ulAliveAddr;
    remote.sin_port = pParam->usAlivePort;

    nRecvSize = recvfrom(nSkt, &buf, MAX_BUF_SIZE, 0, (struct sockaddr *)&remote, &socklength);
    if (nRecvSize < 0)
    {
        printf("[packetParser] recvfrom error [%d]\n", nRecvSize);
    }
    printf("[packetParser] nRecvSize[%d]\n", nRecvSize);

    // Compare WakeUp pattern
    if(isWakeupPattern(buf, pParam->pWakeupPattern, pParam->nWakeupPatternLength))
    {
        if (pParam->nWakeupResponseLength != 0) {
            printf("[packetParser] send wakeup response\n");
            sendWakeupResponse(nSkt, pParam->pWakeupResponse, pParam->nWakeupResponseLength);
        }
        g_bWakeUp = true;
    }
}


static void *thread_WakeUpTask(void *arg)
{
    fd_set readfds;
    struct timeval timeout;
    WakeupParam *pParam = (WakeupParam*)arg;

    int nSkt = pParam->nSkt;

    printf("[thread_WakeUpTask] - skt = %d\n", nSkt);

    int nVal = 0;
    int nMaxFds = nSkt + 1;
    while(1)
    {    
        if(!g_bConnected){
            sleep(1);
            continue;
        }
        SetupSocketSelectParam(nSkt, &readfds, NULL, NULL, &timeout, 1);
        nVal = select(nMaxFds, &readfds, NULL, NULL, &timeout);
        if(nVal > 0)
        {
            if(FD_ISSET(nSkt, &readfds))
            {
                //printf("[thread_WakeUpTask] - FD_ISSET(pParam->nSkt, &readfds), skt = %d\n", nSkt);
                packetParser(pParam);
                //break;
            }
        } else if(nVal == 0) {
            //printf("[thread_WakeUpTask] - select timeout, skt = %d\n", nSkt);
        }

        if(g_bWakeUp)
        {
            printf("[thread_WakeUpTask] - Device is WakeUp!\n");
            break;
        }
    }

    printf("[testIOTC_Wakup] - thread_WakeUpTask exit!\n");
    pthread_exit(0);
}


static void *_thread_Login(void *arg)
{
    int nRet;
    char *UID = (char *)arg;

    while(1)
    {
        nRet = IOTC_Device_Login(UID, "AAAA0009", "12345678");
        printf("Calling IOTC_Device_Login() ret = %d\n", nRet);

        if(nRet == IOTC_ER_NoERROR)
            break;
        else
            sleep(2);
    }
    g_deviceLogined = true;
    pthread_exit(0);
}


static void sendKeepAlivePacket(int nSkt, IOTC_WakeUpDataEx *pIOTCAliveData, int nDataCnt)
{

    if(gSleepProtocol == PROTOCOL_UDP){
        for(int i = 0; i < nDataCnt; i++)
        {
            struct sockaddr_in remote;
            memset(&remote, 0, sizeof(struct sockaddr_in));
            remote.sin_family = AF_INET;
            remote.sin_addr.s_addr = pIOTCAliveData[i].ulAliveAddr;
            remote.sin_port = pIOTCAliveData[i].usAlivePort;

            printf("[%s-UDP] - %s : %d, size = %zu\n", __func__, inet_ntoa(remote.sin_addr), ntohs(remote.sin_port), pIOTCAliveData[i].nAlivePacketLength);

            int ret = sendto(nSkt, pIOTCAliveData[i].pszAlivePacket, pIOTCAliveData[i].nAlivePacketLength, 0, (struct sockaddr *)&remote, sizeof(struct sockaddr_in));
            if (ret < 0)
            {
                printf("[%s-UDP] sendto error\n", __func__);
            }
        }
    }else if(gSleepProtocol == PROTOCOL_TCP){
        printf("[%s-TCP], size = %zu\n", __func__, pIOTCAliveData[0].nAlivePacketLength);

        int ret = send(nSkt, pIOTCAliveData[0].pszAlivePacket, pIOTCAliveData[0].nAlivePacketLength, 0);
        if (ret < 0)
        {
            printf("[%s-TCP] sendto error\n", __func__);
        }
    }
    // sleep(pIOTCAliveData[0].nAliveInterval);
    sleep(3);
}

static void sendWakeupResponse(int nSkt, char *msg, int len) {

    if (gSleepProtocol == PROTOCOL_UDP) {
        for (int i = 0; i < gDataCnt; i++) {
            struct sockaddr_in remote;
            memset(&remote, 0, sizeof(struct sockaddr_in));
            remote.sin_family = AF_INET;
            remote.sin_addr.s_addr = gAliveData[i].ulAliveAddr;
            remote.sin_port = gAliveData[i].usAlivePort;

            printf("[%s-UDP] - %s : %d, size = %d\n", __func__, inet_ntoa(remote.sin_addr), ntohs(remote.sin_port), len);

            int ret = sendto(nSkt, msg, len, 0, (struct sockaddr *)&remote, sizeof(struct sockaddr_in));
            if (ret < 0) {
                printf("[%s-UDP] sendto error\n", __func__);
            }
        }
    } else if (gSleepProtocol == PROTOCOL_TCP) {
        printf("[%s-TCP], size = %d\n", __func__, len);

        int ret = send(nSkt, msg, len, 0);
        if (ret < 0) {
            printf("[%s-TCP] sendto error\n", __func__);
        }
    }
}

static void duplicateWakeUpData(IOTC_WakeUpDataEx *pDest, IOTC_WakeUpDataEx* pSrc, int nDataCnt)
{
    if(pDest && pSrc)
    {
        for(int i = 0; i < nDataCnt; i++)
        {
            memcpy(&pDest[i], &pSrc[i], sizeof(IOTC_WakeUpDataEx));

            // copy Alive packet
            pDest[i].pszAlivePacket = calloc(1, pSrc[i].nAlivePacketLength);
            memcpy(pDest[i].pszAlivePacket, pSrc[i].pszAlivePacket, pSrc[i].nAlivePacketLength);

            // copy wakeup pattern
            pDest[i].pszWakeupPattern = calloc(1, pSrc[i].nWakeupPatternLength);
            memcpy(pDest[i].pszWakeupPattern, pSrc[i].pszWakeupPattern, pSrc[i].nWakeupPatternLength);

            // copy login pattern
            if (pSrc[i].nLoginPacketLength != 0) {
                pDest[i].pszLoginPacket = calloc(1, pSrc[i].nLoginPacketLength);
                memcpy(pDest[i].pszLoginPacket, pSrc[i].pszLoginPacket, pSrc[i].nLoginPacketLength);
            }
        }
    }
}


static void releaseWakeUpData(IOTC_WakeUpDataEx *pData, int nDataCnt)
{
    if(pData)
    {
        for(int i = 0; i < nDataCnt; i++)
        {
            free(pData[i].pszAlivePacket);
            free(pData[i].pszWakeupPattern);
            if (pData[i].nLoginPacketLength != 0) {
                free(pData[i].pszLoginPacket);
            }
        }
        free(pData);
    }
}


static void *thread_AliveTask(void *arg)
{
    AliveParam *ap = (AliveParam*)arg;

    if(gSleepProtocol == PROTOCOL_TCP){
        struct sockaddr_in remote;
        memset(&remote, 0, sizeof(struct sockaddr_in));
        remote.sin_family = AF_INET;
        remote.sin_addr.s_addr = ap->pWUData[0].ulAliveAddr;
        remote.sin_port = htons(80);

        //connect first server
        int ret = connect(ap->nSkt, (struct sockaddr *)&remote, sizeof(struct sockaddr_in));
        printf("connect reseult %d\n", ret);
        if(ret != -1){
            g_bConnected = true;
        }else{
            printf("TCP connect fail %d\n", ret);
        }
    }
    

    while(1)
    {
        if(!g_bWakeUp)
            sendKeepAlivePacket(ap->nSkt, ap->pWUData, ap->nDataCnt);
        else
            break;	
    }

    pthread_exit(0);
}


int IsDataReady(void)
{
    if (gError) {
        return ERROR;
    }
    return gAliveData ? READY : NOT_READY;
}

void SleepPacketCB(IOTC_WakeUpDataEx *pData, unsigned int DataCnt, int nErrCode)
{
    printf("[%s] err [%d], cnt [%d]\n", __func__, nErrCode, DataCnt);
    if (DataCnt > 0 && pData[0].nLoginPacketLength == 0) {
        printf("To FETCH login packet, please execute\"export IOTC_ENABLE_WAKEUP_LOGIN=true\"\n");
    }
    gDataCnt = DataCnt;
    gAliveData = pData;
    gError = nErrCode;
}


int main(int argc, char *argv[])
{
    // ++++++ Devicde at normal mode ++++++
bootup:
    if(argc < 3)
    {
        printf("Argument error!!!\n");
        printf("IOTCAPIs_Device <UID> <protocol type>\n");
        printf("protocol type : UDP or TCP\n");
        return 0;
    }

    printf("IOTCAPIs_Device start...\n");

    int nRet;
    char *UID = (char *)argv[1];

    if (strcmp("TCP", argv[2]) == 0) {
        gSleepProtocol = PROTOCOL_TCP;
        g_bConnected = false;
    } else if (strcmp("UDP", argv[2]) == 0) {
        gSleepProtocol = PROTOCOL_UDP;
    } else {
        printf("Unkonw protocol type!!!\n");
        return -1;
    }

    const char *license_key = getenv(ENV_LICENSE_KEY);
    if (license_key == NULL) {
        printf("Environment variable %s must be set!!\n", ENV_LICENSE_KEY);
        return -1;
    }
    nRet = TUTK_SDK_Set_License_Key(license_key);
    if (nRet != TUTK_ER_NoERROR)
    {
        printf("TUTK_SDK_Set_License_Key()=%d ,exit...!!\n", nRet);
        return 0;
    }

    nRet = IOTC_Initialize2(0);

    printf("Step 1: IOTC_Initialize2() ret = %d\n", nRet);
    if(nRet != IOTC_ER_NoERROR)
    {
        printf("IOTCAPIs_Device exit...!!\n");
        return 0;
    }

    // Use Wake up feature
    IOTC_WakeUp_Init();
    // Login
    pthread_t threadID_Login;
    pthread_create(&threadID_Login, NULL, &_thread_Login, (void *)UID);
    pthread_detach(threadID_Login);

    sleep(2);

    int nSID = -1;
    bool bSleepCondition = true;    // should be replaced by your sleep condition
    while(1)
    {
        if (g_deviceLogined && bSleepCondition)
        {
            //IOTC_WakeUp_Get_SleepPacketEx "Must" be used after login
            gError = 0;
            nRet = IOTC_WakeUp_Get_SleepPacketEx(SleepPacketCB, 0);
            if (nRet == IOTC_ER_NoERROR)
            {
                //wait until data is ready
                while(IsDataReady() == NOT_READY)
                {
                    sleep(1);
                }
                if (IsDataReady() == READY)
                {
                    break;
                }
            }
        }
        printf("Device prepare to listen!\n");
        nSID = IOTC_Listen(10000);
        if (nSID > -1) {
            printf("Device accept connection\n");
        }
    }

    // Device is going to sleep, duplicate IOTC_WakeUpData
    IOTC_WakeUpDataEx *pIOTCAliveData = calloc(1, sizeof(IOTC_WakeUpDataEx) * gDataCnt);
    if(pIOTCAliveData)
        duplicateWakeUpData(pIOTCAliveData, gAliveData, gDataCnt);
    else
        printf("malloc error!\n");

    // Free gAliveData memory
    IOTC_WakeUp_DeInitEx(gAliveData);
    gAliveData = NULL;
    // DeInit IOTC
    IOTC_DeInitialize();
    // ------ Devicde at normal mode ------


    // ++++++ Device enter sleep mode ++++++
    printf("Device enter sleep mode!\n");
    // Driver will create a receive thread and keep receive data
    int nSkt = createSocket();

    WakeupParam *pParam = (WakeupParam*)calloc(1, sizeof(WakeupParam) * gDataCnt);

    for(int i = 0; i < gDataCnt; i++)
    {
        pthread_t nTaskId;

        pParam[i].nSkt = nSkt;
        pParam[i].ulAliveAddr = pIOTCAliveData[i].ulAliveAddr;
        pParam[i].usAlivePort = pIOTCAliveData[i].usAlivePort;
        pParam[i].pWakeupPattern = pIOTCAliveData[i].pszWakeupPattern;
        pParam[i].nWakeupPatternLength = pIOTCAliveData[i].nWakeupPatternLength;
        pParam[i].pWakeupResponse = pIOTCAliveData[i].pszLoginPacket;
        pParam[i].nWakeupResponseLength = pIOTCAliveData[i].nLoginPacketLength;

        nRet = pthread_create(&nTaskId, NULL, thread_WakeUpTask, &pParam[i]);

        if(nRet != 0)
            printf("pthread_create for fail!\n");
    }

    // Driver will create a send thread and keep send IOTC login
    pthread_t nAliveTaskId;

    AliveParam ap;
    ap.nSkt = nSkt;
    ap.pWUData = pIOTCAliveData;
    ap.nDataCnt = gDataCnt;

    nRet = pthread_create(&nAliveTaskId, NULL, thread_AliveTask, &ap);

    pthread_join(nAliveTaskId, 0);

    // Device is woken up
    close(nSkt);
    // Don't forget to free memory
    free(pParam);
    releaseWakeUpData(pIOTCAliveData, gDataCnt);
    // ------ Device enter sleep mode ------

    printf("Device enter normal mode!\n");
    g_bWakeUp = false;
    g_deviceLogined = false;
    goto bootup;
}
